from __future__ import print_function
import argparse
import wandb
import torch
import pandas as pd
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms
from trades import trades_loss


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4 * 4 * 50, 500)
        self.fc2 = nn.Linear(500, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4 * 4 * 50)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


def train(args, model, device, train_loader, optimizer, epoch):
    model.train()
    metrics = pd.DataFrame()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)

        optimizer.zero_grad()

        # calculate robust loss
        loss, batch_metrics = trades_loss(model=model,
                           x_natural=data,
                           y=target,
                           optimizer=optimizer,
                           step_size=args.step_size,
                           epsilon=args.epsilon,
                           perturb_steps=args.num_steps,
                           beta=args.beta,
                           distance='l_inf', args=args)
        metrics = metrics.append(pd.DataFrame(batch_metrics, index=[0]), ignore_index=True)
        loss.backward()
        optimizer.step()
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                       100. * batch_idx / len(train_loader), loss.item()))
    return dict(metrics.mean())


def test(args, model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
        test_loss, correct, len(test_loader.dataset),
        100. * correct / len(test_loader.dataset)))
    return test_loss, correct / len(test_loader.dataset)


def main():
    # Training settings
    parser = argparse.ArgumentParser(description='PyTorch MNIST Example')
    parser.add_argument('--batch-size', type=int, default=128, metavar='N',
                        help='input batch size for training (default: 64)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--epochs', type=int, default=50, metavar='N',
                        help='number of epochs to train (default: 10)')
    parser.add_argument('--lr', type=float, default=0.01, metavar='LR',
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                        help='SGD momentum (default: 0.5)')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--epsilon', default=0.1,
                        help='perturbation')
    parser.add_argument('--num-steps', default=20,
                        help='perturb number of steps')
    parser.add_argument('--step-size', default=0.01,
                        help='perturb step size')
    parser.add_argument('--beta', default=1.0, type=float,
                        help='regularization, i.e., 1/lambda in TRADES')
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                        help='how many batches to wait before logging training status')

    parser.add_argument('--save-model', action='store_true', default=False,
                        help='For Saving the current Model')
    parser.add_argument('--layer_snr_weight_type', default="exp", type=str,
                    help='type of layer snr weight')
    parser.add_argument('--snr_type', default="n2sv", type=str,
                    help='type of snr loss')
    parser.add_argument('--base', default=1.5, type=float,
                            help='base number of ')
    parser.add_argument('--use_snr', action='store_true',
                        help='use snr loss or not')
    parser.add_argument('--snr_layers', default=None, nargs='+', type=str,
                            help='the layers need to compute snr')
    parser.add_argument('--snr_weight', default=10, type=float,
                        help='weight of snr loss')
    parser.add_argument('--save_path', default="", type=str,
                        help='weight of snr loss')
    parser.add_argument('--name', default="", type=str,
                        help='weight of snr loss')
    args = parser.parse_args()
    wandb.init(project="unsupervised-robust-learning", name=args.name)
    use_cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if use_cuda else "cpu")

    kwargs = {'num_workers': 1, 'pin_memory': True} if use_cuda else {}
    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST('../data', train=True, download=True,
                       transform=transforms.Compose([
                           transforms.ToTensor(),
                       ])),
        batch_size=args.batch_size, shuffle=True, **kwargs)
    test_loader = torch.utils.data.DataLoader(
        datasets.MNIST('../data', train=False, transform=transforms.Compose([
            transforms.ToTensor(),
        ])),
        batch_size=args.test_batch_size, shuffle=True, **kwargs)

    model = Net().to(device)
    wandb.watch(model)
    optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    for epoch in range(1, args.epochs + 1):
        res = train(args, model, device, train_loader, optimizer, epoch)
        test_loss, test_clear_acc = test(args, model, device, test_loader)

        if args.use_snr:
            wandb.log({
                        "epoch": epoch,
                        "lr": optimizer.param_groups[0]['lr'],
                        "train_clean_acc" : res['clean_acc'],
                        "test_clean_acc": test_clear_acc,
                        "train_adversarial_acc" : res['adversarial_acc'],
                        "loss":res['loss'],
                        "loss_natural":res['loss_natural'],
                        "loss_robust":res['loss_robust'],
                        "snr_loss":res['snr_loss'],
                        'layer2_snr':res['layer1_snr'],
                        'layer3_snr':res['layer2_snr'],
                        'layer4_snr':res['layer3_snr'],
                        })
        else:
            wandb.log({
                        "epoch": epoch,
                        "lr": optimizer.param_groups[0]['lr'],
                        "train_clean_acc" : res['clean_acc'],
                        "test_clean_acc": test_clear_acc,
                        "train_adversarial_acc" : res['adversarial_acc'],
                        "loss":res['loss'],
                        "loss_natural":res['loss_natural'],
                        "loss_robust":res['loss_robust']
                        })
        
    if (args.save_model):
        torch.save(model.state_dict(), args.save_path)

  
if __name__ == '__main__':
    main()
